import boto3
import os
from .settings import Environment


environment = Environment.production

if os.environ.get('TARGET_ENV_MODE') == 'staging':
    environment = Environment.staging
if os.environ.get('TARGET_ENV_MODE') == 'development':
    environment = Environment.localhost

autodl_host = environment.value['host']
s3_host = environment.value['s3_host']
s3_bucket_name = environment.value['s3_bucket_name']

AUTODL_S3_REGION = 'us-east-1'

# TODO - we need to somehow include these in CLI without hardcoding
# AWS_ACCESS_KEY_ID = os.environ['AWS_ACCESS_KEY_ID']
# AWS_SECRET_ACCESS_KEY = os.environ['AWS_SECRET_ACCESS_KEY']

AWS_ACCESS_KEY_ID = 'AKIASO72NKUYW7ONNRFO'
AWS_SECRET_ACCESS_KEY = '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'

def init_s3():
    return boto3.resource('s3',
        region_name=AUTODL_S3_REGION,
        endpoint_url=s3_host,
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
    )

s3 = init_s3()

def use_environment(new_environment):
    global autodl_host, environment, s3_host, s3_bucket_name, s3
    environment = new_environment

    autodl_host = environment.value['host']
    s3_host = environment.value['s3_host']
    s3_bucket_name = environment.value['s3_bucket_name']
    s3 = init_s3()



import autodl
from click import ClickException
from pathlib import Path
import importlib
import os
import sys

sys.path.append(os.getcwd())

PYTHON_FILE_EXTENSION = '.py'

def analyze_pytorch_source(model_file_path, model_script_args=[]):
    # spec = importlib.util.spec_from_file_location("__main__", model_file_path)
    # module = importlib.util.module_from_spec(spec)
    # spec.loader.exec_module(module)
    args = sys.argv
    sys.argv = [args[0]]
    importlib.import_module(model_file_path[:-len(PYTHON_FILE_EXTENSION)]) # TODO relative imports not working
    sys.argv = [args, model_script_args]
    # exec(open(model_file_path).read(), {})
    models = autodl.attached_models

    if (len(models) < 1):
        # TODO try to manually find the model without autodl attachments
        raise ClickException("The provided model file does not contain any models attached with `autodl.attach_model`.\n"
            + "Please refer to our documentation to read on how to attached your models."
        )
    if (len(models) > 1):
        print ("WARNING: Our current CLI implementation does not support multiple models in one file yet. Please contact us and file an inquiry to add this feature if you need it.")

    model, dummy_input, interns, externs, max_search_depth = models[-1]
    model_file_bytes = autodl.export_pytorch_model_repr(model, dummy_input, interns, externs, max_search_depth)

    return model_file_bytes, dummy_input

def load_tensorflow_protobuf(model_file_path):
    import tensorflow.compat.v1 as tf
    from tensorflow.python.platform import gfile
    from tensorflow.core.protobuf import saved_model_pb2
    from google.protobuf.message import DecodeError
    from tensorflow.python.util import compat

    graph_def = tf.GraphDef()

    with tf.Session() as sess:
        try:
            # this load method cannot load models from https://tfhub.dev/
            # crashes with `google.protobuf.message.DecodeError: Error parsing message with type 'tensorflow.GraphDef'`

            with tf.gfile.GFile(model_file_path, "rb") as f:
                graph_def = tf.GraphDef()
                graph_def.ParseFromString(f.read())

                return [ n for n in graph_def.node ]

        except DecodeError:
            # this load method cannot load models from @marcink
            # nasnet is missing input dimensions
            # posenet is missing meta_graphs - crashes with `IndexError: list index (0) out of range` on tf.import_graph_def(sm.meta_graphs[0].graph_def)
            with gfile.GFile(model_file_path, 'rb') as f:
                data = compat.as_bytes(f.read())
                sm = saved_model_pb2.SavedModel()
                sm.ParseFromString(data)
                tf.import_graph_def(sm.meta_graphs[0].graph_def)
                return [ n for n in sm.meta_graphs[0].graph_def.node ]

def infer_tensorflow_protobuf_input_shape(model_file_path):
    graph_nodes = load_tensorflow_protobuf(model_file_path)
    inferred_batch_size, inferred_width, inferred_height, inferred_channels = None, None, None, None

    for node in graph_nodes:
        if node.op == "Placeholder":
            shape_proto = node.attr['shape']
            dim_sizes = [dim.size for dim in shape_proto.shape.dim]
            print('Detected input layer with shape', dim_sizes)
            dim_sizes += [None, None, None, None]
            dim_sizes = dim_sizes[0:4]
            dim_sizes = [None if size == -1 else size for size in dim_sizes]

            if len(shape_proto.shape.dim) > 0:
                [inferred_batch_size, inferred_width, inferred_height, inferred_channels] = dim_sizes

    return inferred_batch_size, inferred_width, inferred_height, inferred_channels


def analyze_model_file(model_file_path, model_script_args=[]):
    extension = Path(model_file_path).suffixes[-1]

    output = None, None, None, None, None, None, None, None
    model_file, inferred_model_name, framework, inferred_quantization, inferred_channels, inferred_width, inferred_height, inferred_batch_size = output

    model_file = model_file_path
    inferred_model_name = Path(model_file_path).stem
    inferred_quantization = 'FP32'

    if extension in ['.py']:
        framework = 'PYTORCH'
        model_file_bytes, dummy_input = analyze_pytorch_source(model_file_path, model_script_args)
        model_file_name = f'./{inferred_model_name}.adlpt'
        model_file = open(model_file_name, mode="wb")
        model_file.write(model_file_bytes.getbuffer())
        model_file.close()
        model_file = model_file_name

        [inferred_batch_size, inferred_channels, inferred_width, inferred_height] = dummy_input.shape

    elif extension in ['.mar']:
        framework = 'PYTORCH'
        # TODO analyze .mar file for input shape

    elif extension in ['.pb', '.h5']:
        framework = 'TENSORFLOW'

        [inferred_batch_size, inferred_width, inferred_height, inferred_channels] = infer_tensorflow_protobuf_input_shape(model_file_path)

    elif extension in ['.tflite']: # TODO - add this extension also to the UI autofiller
        framework = 'TFLITE'
        raise ClickException(f"Files with '{extension}' are not supported yet")

    elif extension in ['.pt', '.pth']:
        raise ClickException(f"Files with '{extension}' are not supported - please contact us at support@autumn8.ai to add the support.")

    else:
        raise ClickException(f"Files with '{extension}' are not supported - please contact us at support@autumn8.ai to add the support.")


    return model_file, inferred_model_name, framework, inferred_quantization, inferred_channels, inferred_width, inferred_height, inferred_batch_size



import os
import sys
import torch
import io
import zipfile
import json
import pickle
import requests
import urllib
import uuid
import torchvision

sys.path.append(os.path.join(os.path.dirname(__file__), '../'))
from autodl.package_resolver import package
from common._version import __version__
from common.config.s3 import s3, autodl_host, s3_bucket_name
from .package_resolver import package

# TODO: add support for other organizations to CLI
DEFAULT_ORGANIZATION_ID=1




class LayerConfig(object):
    params = list()
    name = None

    def __init__(self, name, params):
        self.name = name
        self.params = params

    def __str__(self):
        return json.dumps(self.__dict__)

    def __repr__(self):
        return self.__str__()

def get_layers_configs(model, dummy_input):
    layers = get_children_of_model(model)

    input_shapes = {}
    def get_input_shape(name):
        def hook(model, input, output):
            #activation[name] = output.detach()
            try:
                input_shapes[name] = input[0].shape
            except:
                pass
        return hook

    output_shapes = {}
    def get_output_shape(name):
        def hook(model, input, output):
            #activation[name] = output.detach()
            try:
                output_shapes[name] = output.shape
            except:
                pass
        return hook

    for i, layer in enumerate(layers):
        layer.register_forward_hook(get_input_shape(str(i)))
        layer.register_forward_hook(get_output_shape(str(i)))

    model(*dummy_input)

    layers_configs = []

    for i, layer in enumerate(layers):
        try:
            input_shape = input_shapes[str(i)]
            output_shape = output_shapes[str(i)]
        except:
            continue
        if isinstance(layer, torch.nn.Conv2d):
            params = [float(input_shape[1]), float(input_shape[2]), float(layer.out_channels), float(layer.kernel_size[0]), float(layer.stride[0]), float(layer.dilation[0]), float(layer.padding[0]), float(input_shape[0])]
            config = LayerConfig("Conv2D", params)
        elif isinstance(layer, torch.nn.Linear):
            params = [float(input_shape[1]), layer.out_features, float(input_shape[0])]
            config = LayerConfig("Linear", params)
        elif isinstance(layer, torch.nn.ReLU):
            x = input_shape[2] if len(input_shape) > 2 else 1
            params = [float(input_shape[1]), x ,float(input_shape[0])]
            config = LayerConfig("ReLU", params)
        elif isinstance(layer, torch.nn.EmbeddingBag):
            params = [float(input_shape[0]), float(output_shape[0]), float(output_shape[1]), float(layer.num_embeddings)]
            config = LayerConfig("EmbeddingBag", params)
        elif isinstance(layer, torch.nn.Dropout2d):
            x = input_shape[2] if len(input_shape)>2 else 1
            params = [float(input_shape[1]), x, layer.p, float(input_shape[0])]
            config = LayerConfig("Dropout2d", params)
        elif isinstance(layer, torch.nn.Dropout):
            params = [float(input_shape[1]), layer.p, float(input_shape[0])]
            config = LayerConfig("Dropout", params)
        elif isinstance(layer, torchvision.ops.misc.FrozenBatchNorm2d):
            params = [float(input_shape[2]), layer.weight.shape[0], float(input_shape[0])]
            config = LayerConfig("BatchNorm2d", params)
        elif isinstance(layer, torch.nn.BatchNorm2d):
            params = [float(input_shape[2]), layer.num_features, float(input_shape[0])]
            config = LayerConfig("BatchNorm2d", params)
        elif isinstance(layer, torch.nn.MaxPool2d):
            params = [float(input_shape[1]), float(input_shape[2]), float(layer.kernel_size), float(layer.stride), float(layer.dilation), float(layer.padding), float(input_shape[0])]
            config = LayerConfig("MaxPool2d", params)
        elif isinstance(layer, torch.nn.AdaptiveAvgPool2d):
            params = [float(input_shape[1]), float(input_shape[2]), float(input_shape[3]), float(layer.output_size[0])]
            config = LayerConfig("AdaptiveAvgPool2d", params)
        else:
            #unknown layer type
            config = LayerConfig(type(layer).__name__, None)

        layers_configs.append(config)

    return layers_configs


def get_children_of_model(model: torch.nn.Module):
    # get children from model!
    children = list(model.children())
    flatt_children = []
    if children == []:
        # return wrapped in array, it will be flattened and return type will be consistant
        # also protects from error from fully functional models
        return [model]
    else:
       # look for children from children... to the last child!
       for child in children:
            try:
                flatt_children.extend(get_children_of_model(child))
            except TypeError:
                flatt_children.append(get_children_of_model(child))
    return flatt_children

def write_model(model, dummy_input, bytes):
    traced = torch.jit.trace(model, dummy_input)
    traced.save(bytes)

def export_pytorch_model_repr(model, dummy_input, interns = [], externs = [], max_search_depth = 5):
    bytes = io.BytesIO()

    assert issubclass(model.__class__, torch.nn.Module) # @mbalc - There actually was one PyTorch model I had problems with that didn't inherit from torch Module o.O

    if not isinstance(dummy_input, tuple):
        dummy_input = (dummy_input,)

    file, requirements, package_name = package(model, interns, externs, max_search_depth)

    layers_configs = get_layers_configs(model, dummy_input)

    with zipfile.ZipFile(bytes, "w") as zip:
        zip.writestr("modelConfig", pickle.dumps(layers_configs))
        zip.writestr("MANIFEST", json.dumps({
            "version": __version__,
            "package_name": package_name,
        }))
        zip.write(file.name, arcname="model.package")
        requirement_list = []
        for package_name,package_version in requirements.items():
            requirement_list.append(f"{package_name}=={package_version}")

        zip.writestr("requirements.txt", "\n".join(requirement_list))

    bytes.seek(0)
    return bytes

def load(filename):
    with zipfile.ZipFile(filename) as z:
        with z.open("modelConfig", "r") as modelFile:
            modelConfig = pickle.loads(modelFile.read())

    return modelConfig

def loadModel(filename):
    with zipfile.ZipFile(filename) as z:
        with z.open("MANIFEST", "r") as manifest:
            package_name = json.loads(manifest.read())["package_name"]
        with z.open("model.package", "r") as modelFile:
            resource_name = "model.pkl"
            imp = torch.package.PackageImporter(modelFile)
            loaded_model = imp.load_pickle(package_name, resource_name)
            return loaded_model


def urlWithParams(url, params):
    url_parse = urllib.parse.urlparse(url)
    url_new_query = urllib.parse.urlencode(params)
    url_parse = url_parse._replace(query=url_new_query)

    new_url = urllib.parse.urlunparse(url_parse)
    return new_url

def postModel(model_config):
    api_route = f"{autodl_host}/api/lab/model"
    print('submitting model to', api_route)
    response = requests.post(urlWithParams(api_route, {
        "organization_id": DEFAULT_ORGANIZATION_ID
    }), headers={"Content-Type": "application/json"}, data=json.dumps(model_config))

    if response.status_code != 200:
        raise Exception("Received response {}".format(response.status_code))
    return json.loads(response.text)["model_id"]

def deleteModel(model_id):
    new_url = urlWithParams("{}/api/lab/model".format(autodl_host), {
        "model_id": model_id,
        "organization_id": DEFAULT_ORGANIZATION_ID
    })
    response = requests.delete(new_url)
    if response.status_code != 200:
        raise Exception("Received response {}".format(response.status_code))

def postModelFile(bytes_or_filepath, s3_file_url):
    if isinstance(bytes_or_filepath, io.BytesIO):
        f = bytes_or_filepath
        s3.Bucket(s3_bucket_name).Object(s3_file_url).upload_fileobj(f)
    else:
        with open(bytes_or_filepath, "rb") as f:
            s3.Bucket(s3_bucket_name).Object(s3_file_url).upload_fileobj(f)

def asyncEstimation(model_id):
    new_url = urlWithParams("{}/api/lab/model/async_prediction".format(autodl_host), {
        "model_id": model_id,
        "organization_id": DEFAULT_ORGANIZATION_ID
    })
    response = requests.post(new_url)
    if response.status_code != 200:
        raise Exception("Received response {}".format(response.status_code))
    return response

def uploadModel(model_config, model_file):
    s3_file_url = f"autodl-staging/models/{uuid.uuid4()}-{model_config['name']}"
    if type(model_file) == io.BytesIO:
        model_file.seek(0)

    print('Uploading the model files...')
    postModelFile(model_file, s3_file_url)
    model_config['s3_file_url'] = s3_file_url

    print('Creating the model entry in AutoDL...')
    model_id = postModel(model_config)

    print('Starting up performance predictor...')
    asyncEstimation(model_id)
    return "{}/{}/performancePredictor/dashboard/{}".format(autodl_host, DEFAULT_ORGANIZATION_ID, model_id)

attached_models = []

def attach_model(model, example_input, interns = [], externs = [], max_search_depth = 5):
    attached_models.append((model, example_input, interns, externs, max_search_depth))



import os
from setuptools import setup, find_packages
from common._version import __version__

dependencies = ["wheel>=0.31.0", "boto3>=1.20.15", "tensorflow", "click>=8.0.1", "questionary>=1.10.0", "torch", "torchvision"]

setup(
   name='autumn8',
   version=__version__,
   author='Autumn8',
   author_email='marcink@radcode.co',

   install_requires=dependencies,
   setup_requires=dependencies,


   packages=[
      *find_packages(),
   ],
   description='Utilities to export models to autodl',

   entry_points={
      'console_scripts': [
         'autodl-cli=cli:main',
      ],
   },
)



import json
import click
import random
import questionary

import autodl
from common._version import __version__
from cli.examples import example_model_names
from cli.analyze import analyze_model_file

from common.config.s3 import use_environment
from common.config.settings import supported_quants, supported_domains, supported_batch_sizes, Environment

@click.group()
@click.version_option(version=__version__)
def main():
    pass

# cannot use click prompt kwargs feature for the command options, because we want to infer input dimensions and the model name
def prompt_for_missing_model_info(
    model_name, quantization, domain, task, channels, width, height, batch_size,
    inferred_model_name, inferred_quantization, inferred_channels, inferred_width, inferred_height, inferred_batch_size
):
    # TODO - attempt reading model details from some configCache files
    if model_name is None:
        model_name = questionary.text(
            f"Please give a name to your model to be used in AutoDL (for example: '{random.choice(example_model_names)}')\n  Model name:",
            validate=lambda name: len(name) > 0,
            default=inferred_model_name,
        ).unsafe_ask()
    if quantization is None:
        quantization = questionary.select(
            "Choose quantization for the model",
            choices=supported_quants,
            use_shortcuts=True,
            default=inferred_quantization,
        ).unsafe_ask()
    if domain is None:
        doms = [dom['domain'] for dom in supported_domains]
        domain = questionary.select(
            "Choose problem domain for the model",
            choices=doms,
            use_shortcuts=True,
        ).unsafe_ask()
    if task is None:
        dom_tasks = [task for dom in supported_domains if dom['domain'] == domain for task in dom['tasks']]
        task = questionary.select(
            "Choose task type for the model",
            choices=dom_tasks,
            use_shortcuts=True,
        ).unsafe_ask()
    if channels is None:
        channels = questionary.text(
            "Specify channel count for the model input",
            validate=lambda name: name.isnumeric(),
            default=str(inferred_channels) if inferred_channels else '',
        ).unsafe_ask()
    if width is None:
        width = questionary.text(
            "Specify input width for the model",
            validate=lambda name: name.isnumeric(),
            default=str(inferred_width) if inferred_width else '',
        ).unsafe_ask()
    if height is None:
        height = questionary.text(
            "Specify input height for the model",
            validate=lambda name: name.isnumeric(),
            default=str(inferred_height) if inferred_height else '',
        ).unsafe_ask()
    if batch_size is None:
        batch_size = questionary.select(
            "Specify input batch size for the model",
            choices = [str(bs) for bs in supported_batch_sizes],
            use_shortcuts=True,
            default=str(inferred_batch_size) if inferred_batch_size else None,
        ).unsafe_ask()

    return model_name, quantization, domain, task, channels, width, height, batch_size


@main.command()
@click.argument('model_file_path')
@click.argument('model_script_args', nargs=-1)
@click.option('-n', '--name', 'model_name',
    help="Name of the model to be used in AutoDL.",
)
@click.option('-e', '--environment', '--env',
    type=click.Choice(Environment.__members__, case_sensitive=False),
    default='production',
    callback=lambda c, p, v: getattr(Environment, v),
    help="Environment to upload the model to.",
)
@click.option('-q', '--quantization', '--quants',
    type=click.Choice(supported_quants, case_sensitive=False),
    help="Quantization for the model.",
)
@click.option('-d', '--domain',
    type=click.Choice([dom['domain'] for dom in supported_domains], case_sensitive=False),
    help="Domain of the problem the model is addressing.",
)
@click.option('-t', '--task',
    type=click.Choice([task for dom in supported_domains for task in dom['tasks']], case_sensitive=False),
    help="The task type the model is solving.",
)
@click.option('-c', '--channels',
    type=int,
    help="The channel count for the model input.",
)
@click.option('-w', '--width',
    type=int,
    help="The model input width.",
)
@click.option('-h', '--height',
    type=int,
    help="The model input height.",
)
@click.option('-b', '--batch_size',
    type=int,
    help="The model input batch_size to benchmark for.",
)
@click.option('-y', '--yes', 'auto_confirm', is_flag=True, help="Skip all confirmation input from the user.")
def submit_model(model_file_path,
                model_script_args,
                model_name,
                auto_confirm,
                quantization,
                domain,
                task,
                channels,
                width,
                height,
                batch_size,
                **kwargs):
    # Priority: flags, then configCache, then inference, then interactive user input
    environment = kwargs['environment']
    use_environment(environment)

    model_file, inferred_model_name, framework, inferred_quantization, inferred_channels, inferred_width, inferred_height, inferred_batch_size = analyze_model_file(model_file_path, model_script_args)

    model_name, quantization, domain, task, channels, width, height, batch_size = prompt_for_missing_model_info(
        model_name, quantization, domain, task, channels, width, height, batch_size,
        inferred_model_name, inferred_quantization, inferred_channels, inferred_width, inferred_height, inferred_batch_size
    )
    model_config = {
        "name": model_name,
        "framework": framework,
        "quantization": quantization,
        "from_onspecta": False, # TODO we should disallow this param in the API
        "domain": domain,
        "task": task,
        "channels": channels,
        "height": height,
        "width": width,
        "batch_size": batch_size,
    }

    if not auto_confirm:
        click.confirm(
            text = "\n"
                + "The details for your model are as follows:\n"
                + f"{json.dumps(model_config, indent=4)}\n"
                + "\n"
                + "Are you sure you want to upload that to AutoDL?",
            abort=True,
            default=True
        )

    model_link = autodl.uploadModel(model_config, model_file)
    print(f'Done! See {model_link}')

if __name__ == '__main__':
    main()
